@@ -5,6 +5,9 @@ import os
 import logging
 import glob
 import shutil
+import signal
+import stat
+import sys
 import time
 import platform
 import gbinder
@@ -247,6 +250,10 @@ def make_base_props(args):
     if not os.path.exists("/dev/ashmem"):
         props.append("sys.use_memfd=true")
 
+    # Added for security reasons
+    props.append("ro.adb.secure=1")
+    props.append("ro.debuggable=0")
+
     egl = tools.helpers.props.host_get(args, "ro.hardware.egl")
     dri, _ = tools.helpers.gpu.getDriNode(args)
 
@@ -400,7 +407,10 @@ def start(args):
     tools.helpers.run.user(args, command, output="background")
     wait_for_running(args)
     # Workaround lxc-start changing stdout/stderr permissions to 700
-    os.chmod(args.log, 0o666)
+    try:
+        os.chmod(args.log, 0o666)
+    except:
+        pass
 
 def stop(args):
     command = ["lxc-stop", "-P",
@@ -491,10 +501,18 @@ def shell(args):
     else:
         command.append("/system/bin/sh")
 
+    def mock_sigint(signum, frame):
+        raise KeyboardInterrupt
+    signal.signal(signal.SIGTERM, mock_sigint)
+    signal.signal(signal.SIGHUP, mock_sigint)
+
+    perms = stat.S_IMODE(os.stat(sys.stdout.fileno()).st_mode)
     try:
         subprocess.run(command)
     except KeyboardInterrupt:
         pass
+    finally:
+        os.chmod(sys.stdout.fileno(), perms)
 
     if state == "FROZEN":
         freeze(args)
